from .context import ActionContext
from rid_lib.types import KoiNetNode
from rid_lib.ext import Bundle
from .effector import Effector


@Effector.register_default_action(KoiNetNode)
def dereference_koi_node(ctx: ActionContext, rid: KoiNetNode) -> Bundle:
    if rid != ctx.identity.rid:
        return
    
    return Bundle.generate(
        rid=ctx.identity.rid,
        contents=ctx.identity.profile.model_dump()
    )


import logging
from rich.logging import RichHandler
from pydantic import Field
from rid_lib.types import KoiNetNode, KoiNetEdge
from koi_net.config import NodeConfig, KoiNetConfig
from koi_net.protocol.node import NodeProfile, NodeProvides, NodeType
from koi_net import NodeInterface
from koi_net.context import HandlerContext
from koi_net.processor.handler import HandlerType
from koi_net.processor.knowledge_object import KnowledgeObject
from koi_net.protocol.event import Event, EventType
from koi_net.protocol.edge import EdgeType, generate_edge_bundle

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    handlers=[RichHandler()]
)

logging.getLogger("koi_net").setLevel(logging.DEBUG)
logger = logging.getLogger(__name__)

class CoordinatorConfig(NodeConfig):
    koi_net: KoiNetConfig = Field(default_factory = lambda:
        KoiNetConfig(
            node_name="coordinator",
            node_profile=NodeProfile(
                node_type=NodeType.FULL,
                provides=NodeProvides(
                    event=[KoiNetNode, KoiNetEdge],
                    state=[KoiNetNode, KoiNetEdge]
                )
            ),
            cache_directory_path=".coordinator_rid_cache",
            event_queues_path="coordinator_event_queues.json",
            private_key_pem_path="coordinator_priv_key.pem"
        )
    )
    
node = NodeInterface(
    config=CoordinatorConfig.load_from_yaml("coordinator_config.yaml"),
    use_kobj_processor_thread=True
)

@node.processor.pipeline.register_handler(HandlerType.Network, rid_types=[KoiNetNode])
def handshake_handler(ctx: HandlerContext, kobj: KnowledgeObject):
    logger.info("Handling node handshake")

    # only respond if node declares itself as NEW
    if kobj.event_type != EventType.NEW:
        return
        
    logger.info("Sharing this node's bundle with peer")
    identity_bundle = ctx.effector.deref(ctx.identity.rid)
    ctx.event_queue.push_event_to(
        event=Event.from_bundle(EventType.NEW, identity_bundle),
        node=kobj.rid,
        flush=True
    )
    
    logger.info("Proposing new edge")    
    # defer handling of proposed edge
    
    edge_bundle = generate_edge_bundle(
        source=kobj.rid,
        target=ctx.identity.rid,
        edge_type=EdgeType.WEBHOOK,
        rid_types=[KoiNetNode, KoiNetEdge]
    )
        
    ctx.handle(rid=edge_bundle.rid, event_type=EventType.FORGET)
    ctx.handle(bundle=edge_bundle)
    
if __name__ == "__main__":
    node.server.run()


from logging import getLogger
from koi_net.protocol.errors import ErrorTypes
from koi_net.protocol.event import EventType
from rid_lib.types import KoiNetNode
from ..processor.interface import ProcessorInterface
from ..network.behavior import Actor

logger = getLogger(__name__)


class ErrorHandler:
    timeout_counter: dict[KoiNetNode, int]
    processor: ProcessorInterface
    actor: Actor
    
    def __init__(
        self, 
        processor: ProcessorInterface,
        actor: Actor
    ):
        self.processor = processor
        self.actor = actor
        self.timeout_counter = {}
        
    def handle_connection_error(self, node: KoiNetNode):
        self.timeout_counter.setdefault(node, 0)
        self.timeout_counter[node] += 1
        
        logger.debug(f"{node} has timed out {self.timeout_counter[node]} time(s)")
        
        if self.timeout_counter[node] > 3:
            logger.debug(f"Exceeded time out limit, forgetting node")
            self.processor.handle(rid=node, event_type=EventType.FORGET)
            # do something
        
        
    def handle_protocol_error(
        self, 
        error_type: ErrorTypes, 
        node: KoiNetNode
    ):
        logger.info(f"Handling protocol error {error_type} for node {node!r}")
        match error_type:
            case ErrorTypes.UnknownNode:
                logger.info("Peer doesn't know me, attempting handshake...")
                self.actor.handshake_with(node)
                
            case ErrorTypes.InvalidKey: ...
            case ErrorTypes.InvalidSignature: ...
            case ErrorTypes.InvalidTarget: ...




import time
import logging
from .processor.interface import ProcessorInterface
from .lifecycle import NodeLifecycle
from .network.resolver import NetworkResolver
from .config import NodeConfig

logger = logging.getLogger(__name__)


class NodePoller:
    def __init__(
        self,
        processor: ProcessorInterface,
        lifecycle: NodeLifecycle,
        resolver: NetworkResolver,
        config: NodeConfig
    ):
        self.processor = processor
        self.lifecycle = lifecycle
        self.resolver = resolver
        self.config = config

    def run(self):
        try:
            self.lifecycle.start()
            while True:
                start_time = time.time()
                neighbors = self.resolver.poll_neighbors()
                for node_rid in neighbors:
                    for event in neighbors[node_rid]:
                        self.processor.handle(event=event, source=node_rid)
                self.processor.flush_kobj_queue()
                
                elapsed = time.time() - start_time
                sleep_time = self.config.koi_net.polling_interval - elapsed
                if sleep_time > 0:
                    time.sleep(sleep_time)
                    
        except KeyboardInterrupt:
            logger.info("Polling interrupted by user.")
            
        finally:
            self.lifecycle.stop()


import logging
import uvicorn
from contextlib import asynccontextmanager
from fastapi import FastAPI, APIRouter
from fastapi.responses import JSONResponse
from .network.event_queue import NetworkEventQueue
from .network.response_handler import ResponseHandler
from .processor.interface import ProcessorInterface
from .protocol.api_models import (
    PollEvents,
    FetchRids,
    FetchManifests,
    FetchBundles,
    EventsPayload,
    RidsPayload,
    ManifestsPayload,
    BundlesPayload,
    ErrorResponse
)
from .protocol.errors import ProtocolError
from .protocol.envelope import SignedEnvelope
from .protocol.consts import (
    BROADCAST_EVENTS_PATH,
    POLL_EVENTS_PATH,
    FETCH_RIDS_PATH,
    FETCH_MANIFESTS_PATH,
    FETCH_BUNDLES_PATH
)
from .secure import Secure
from .lifecycle import NodeLifecycle
from .config import NodeConfig

logger = logging.getLogger(__name__)


class NodeServer:
    lifecycle: NodeLifecycle
    
    def __init__(
        self,
        config: NodeConfig,
        lifecycle: NodeLifecycle,
        secure: Secure,
        processor: ProcessorInterface,
        event_queue: NetworkEventQueue,
        response_handler: ResponseHandler
    ):
        self.config = config
        self.lifecycle = lifecycle
        self.secure = secure
        self.processor = processor
        self.event_queue = event_queue
        self.response_handler = response_handler
        self._build_app()
        
    def _build_app(self):
        self.app = FastAPI(
            lifespan=self.lifespan, 
            title="KOI-net Protocol API",
            version="1.0.0"
        )
        
        self.router = APIRouter(prefix="/koi-net")
        self.app.add_exception_handler(ProtocolError, self.protocol_error_handler)
        
        def _add_endpoint(path, func):
            self.router.add_api_route(
                path=path,
                endpoint=self.secure.envelope_handler(func),
                methods=["POST"]
            )
        
        _add_endpoint(BROADCAST_EVENTS_PATH, self.broadcast_events)
        _add_endpoint(POLL_EVENTS_PATH, self.poll_events)
        _add_endpoint(FETCH_RIDS_PATH, self.fetch_rids)
        _add_endpoint(FETCH_MANIFESTS_PATH, self.fetch_manifests)
        _add_endpoint(FETCH_BUNDLES_PATH, self.fetch_bundles)
        
        self.app.include_router(self.router)
    
    def run(self):
        uvicorn.run(
            app=self.app,
            host=self.config.server.host,
            port=self.config.server.port
        )
        
    @asynccontextmanager
    async def lifespan(self, app: FastAPI):
        self.lifecycle.start()
        yield
        self.lifecycle.stop()

    def protocol_error_handler(self, request, exc: ProtocolError):
        logger.info(f"caught protocol error: {exc}")
        resp = ErrorResponse(error=exc.error_type)
        logger.info(f"returning error response: {resp}")
        return JSONResponse(
            status_code=400,
            content=resp.model_dump(mode="json")
        )

    async def broadcast_events(self, req: SignedEnvelope[EventsPayload]):
        logger.info(f"Request to {BROADCAST_EVENTS_PATH}, received {len(req.payload.events)} event(s)")
        for event in req.payload.events:
            self.processor.handle(event=event, source=req.source_node)
        
    async def poll_events(
        self, req: SignedEnvelope[PollEvents]
    ) -> SignedEnvelope[EventsPayload] | ErrorResponse:
        logger.info(f"Request to {POLL_EVENTS_PATH}")
        events = self.event_queue.flush_poll_queue(req.payload.rid)
        return EventsPayload(events=events)

    async def fetch_rids(
        self, req: SignedEnvelope[FetchRids]
    ) -> SignedEnvelope[RidsPayload] | ErrorResponse:
        return self.response_handler.fetch_rids(req.payload)

    async def fetch_manifests(
        self, req: SignedEnvelope[FetchManifests]
    ) -> SignedEnvelope[ManifestsPayload] | ErrorResponse:
        return self.response_handler.fetch_manifests(req.payload)

    async def fetch_bundles(
        self, req: SignedEnvelope[FetchBundles]
    ) -> SignedEnvelope[BundlesPayload] | ErrorResponse:
        return self.response_handler.fetch_bundles(req.payload)



import logging
from koi_net.protocol.node import NodeType
from rid_lib.ext import Cache
from .network.resolver import NetworkResolver
from .network.event_queue import NetworkEventQueue
from .network.graph import NetworkGraph
from .network.request_handler import RequestHandler
from .network.response_handler import ResponseHandler
from .network.error_handler import ErrorHandler
from .network.behavior import Actor
from .processor.interface import ProcessorInterface
from .processor import default_handlers
from .processor.handler import KnowledgeHandler
from .processor.knowledge_pipeline import KnowledgePipeline
from .identity import NodeIdentity
from .secure import Secure
from .config import NodeConfig
from .context import HandlerContext, ActionContext
from .effector import Effector
from .server import NodeServer
from .lifecycle import NodeLifecycle
from .poller import NodePoller
from . import default_actions

logger = logging.getLogger(__name__)



class NodeInterface:
    config: NodeConfig
    cache: Cache
    identity: NodeIdentity
    resolver: NetworkResolver
    event_queue: NetworkEventQueue
    graph: NetworkGraph
    processor: ProcessorInterface
    secure: Secure
    server: NodeServer
    
    use_kobj_processor_thread: bool
    
    def __init__(
        self,
        config: NodeConfig,
        use_kobj_processor_thread: bool = False,
        handlers: list[KnowledgeHandler] | None = None,
        cache: Cache | None = None,
        processor: ProcessorInterface | None = None
    ):
        self.config = config
        self.cache = cache or Cache(
            directory_path=self.config.koi_net.cache_directory_path
        )

        self.identity = NodeIdentity(config=self.config)
        self.effector = Effector(cache=self.cache)

        self.graph = NetworkGraph(
            cache=self.cache, 
            identity=self.identity
        )
        
        self.secure = Secure(
            identity=self.identity, 
            effector=self.effector, 
            config=self.config
        )
        
        self.request_handler = RequestHandler(
            effector=self.effector, 
            identity=self.identity,
            secure=self.secure
        )
        
        self.response_handler = ResponseHandler(self.cache, self.effector)
        
        self.resolver = NetworkResolver(
            config=self.config,
            cache=self.cache, 
            identity=self.identity,
            graph=self.graph,
            request_handler=self.request_handler,
            effector=self.effector
        )
        
        self.event_queue = NetworkEventQueue(
            config=self.config,
            cache=self.cache, 
            identity=self.identity,
            graph=self.graph,
            request_handler=self.request_handler,
            effector=self.effector
        )
        
        self.actor = Actor(
            identity=self.identity,
            effector=self.effector,
            event_queue=self.event_queue
        )
        
        # pull all handlers defined in default_handlers module
        if handlers is None:
            handlers = [
                obj for obj in vars(default_handlers).values() 
                if isinstance(obj, KnowledgeHandler)
            ]

        self.use_kobj_processor_thread = use_kobj_processor_thread
        
        self.action_context = ActionContext(
            identity=self.identity,
            effector=self.effector
        )
        
        self.handler_context = HandlerContext(
            identity=self.identity,
            cache=self.cache,
            event_queue=self.event_queue,
            graph=self.graph,
            request_handler=self.request_handler,
            effector=self.effector
        )
        
        self.pipeline = KnowledgePipeline(
            handler_context=self.handler_context,
            cache=self.cache,
            request_handler=self.request_handler,
            event_queue=self.event_queue,
            graph=self.graph,
            default_handlers=handlers
        )
        
        self.processor = processor or ProcessorInterface(
            pipeline=self.pipeline,
            use_kobj_processor_thread=self.use_kobj_processor_thread
        )
        
        self.error_handler = ErrorHandler(
            processor=self.processor,
            actor=self.actor
        )
        
        self.request_handler.set_error_handler(self.error_handler)
        
        self.handler_context.set_processor(self.processor)
        
        self.effector.set_processor(self.processor)
        self.effector.set_resolver(self.resolver)
        self.effector.set_action_context(self.action_context)
        
        self.lifecycle = NodeLifecycle(
            config=self.config,
            identity=self.identity,
            graph=self.graph,
            processor=self.processor,
            effector=self.effector,
            actor=self.actor,
            use_kobj_processor_thread=use_kobj_processor_thread
        )
        
        # if self.config.koi_net.node_profile.node_type == NodeType.FULL:
        self.server = NodeServer(
            config=self.config,
            lifecycle=self.lifecycle,
            secure=self.secure,
            processor=self.processor,
            event_queue=self.event_queue,
            response_handler=self.response_handler
        )
        
        self.poller = NodePoller(
            processor=self.processor,
            lifecycle=self.lifecycle,
            resolver=self.resolver,
            config=self.config
        )



import logging
from typing import Callable
from enum import StrEnum
from rid_lib.ext import Cache, Bundle
from rid_lib.core import RID, RIDType
from rid_lib.types import KoiNetNode

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .network.resolver import NetworkResolver
    from .processor.interface import ProcessorInterface
    from .context import ActionContext

logger = logging.getLogger(__name__)


class BundleSource(StrEnum):
    CACHE = "CACHE"
    ACTION = "ACTION"

class Effector:
    cache: Cache
    resolver: "NetworkResolver | None"
    processor: "ProcessorInterface | None"
    action_context: "ActionContext | None"
    _action_table: dict[
        type[RID], 
        Callable[
            ["ActionContext", RID], 
            Bundle | None
        ]
    ] = dict()
    
    def __init__(
        self, 
        cache: Cache,
    ):
        self.cache = cache
        self.resolver = None
        self.processor = None
        self.action_context = None
        self._action_table = self.__class__._action_table.copy()
        
    def set_processor(self, processor: "ProcessorInterface"):
        self.processor = processor
        
    def set_resolver(self, resolver: "NetworkResolver"):
        self.resolver = resolver
        
    def set_action_context(self, action_context: "ActionContext"):
        self.action_context = action_context
        
    @classmethod
    def register_default_action(cls, rid_type: RIDType):
        def decorator(func: Callable) -> Callable:
            cls._action_table[rid_type] = func
            return func
        return decorator
        
    def register_action(self, rid_type: RIDType):
        def decorator(func: Callable) -> Callable:
            self._action_table[rid_type] = func
            return func
        return decorator
    
    def _try_cache(self, rid: RID) -> tuple[Bundle, BundleSource] | None:
        bundle = self.cache.read(rid)
        
        if bundle:
            logger.debug("Cache hit")
            return bundle, BundleSource.CACHE
        else:
            logger.debug("Cache miss")
            return None
                    
    def _try_action(self, rid: RID) -> tuple[Bundle, BundleSource] | None:
        if type(rid) not in self._action_table:
            logger.debug("No action available")
            return None
        
        logger.debug("Action available")
        func = self._action_table[type(rid)]
        bundle = func(
            ctx=self.action_context, 
            rid=rid
        )
        
        if bundle:
            logger.debug("Action hit")
            return bundle, BundleSource.ACTION
        else:
            logger.debug("Action miss")
            return None

        
    def _try_network(self, rid: RID) -> tuple[Bundle, KoiNetNode] | None:
        bundle, source = self.resolver.fetch_remote_bundle(rid)
        
        if bundle:
            logger.debug("Network hit")
            return bundle, source
        else:
            logger.debug("Network miss")
            return None
        
    
    def deref(
        self, 
        rid: RID,
        refresh_cache: bool = False,
        use_network: bool = False,
        handle_result: bool = True
    ) -> Bundle | None:
        logger.debug(f"Dereferencing {rid!r}")
        
        bundle, source = (
            # if `refresh_cache`, skip try cache
            not refresh_cache and self._try_cache(rid) or 
            self._try_action(rid) or
            use_network and self._try_network(rid) or
            # if not found, bundle and source set to None
            (None, None) 
        )
        
        if (
            handle_result 
            and bundle is not None 
            and source != BundleSource.CACHE
        ):            
            self.processor.handle(
                bundle=bundle, 
                source=source if type(source) is KoiNetNode else None
            )

            # TODO: refactor for general solution, param to write through to cache before continuing
            # like `self.processor.kobj_queue.join()``

        return bundle


import logging
import queue
import threading
from rid_lib.core import RID
from rid_lib.ext import Bundle, Manifest
from rid_lib.types import KoiNetNode
from ..protocol.event import Event, EventType
from .knowledge_object import KnowledgeObject
from .knowledge_pipeline import KnowledgePipeline


logger = logging.getLogger(__name__)


class ProcessorInterface:
    """Provides access to this node's knowledge processing pipeline."""
    pipeline: KnowledgePipeline
    kobj_queue: queue.Queue[KnowledgeObject]
    use_kobj_processor_thread: bool
    worker_thread: threading.Thread | None = None
    
    def __init__(
        self,
        pipeline: KnowledgePipeline,
        use_kobj_processor_thread: bool,
    ):
        self.pipeline = pipeline
        self.use_kobj_processor_thread = use_kobj_processor_thread
        self.kobj_queue = queue.Queue()
        
        if self.use_kobj_processor_thread:
            self.worker_thread = threading.Thread(
                target=self.kobj_processor_worker,
                daemon=True
            )
    
    def flush_kobj_queue(self):
        """Flushes all knowledge objects from queue and processes them.
        
        NOTE: ONLY CALL THIS METHOD IN SINGLE THREADED NODES, OTHERWISE THIS WILL CAUSE RACE CONDITIONS.
        """
        if self.use_kobj_processor_thread:
            logger.warning("You are using a worker thread, calling this method can cause race conditions!")
        
        while not self.kobj_queue.empty():
            kobj = self.kobj_queue.get()
            logger.debug(f"Dequeued {kobj!r}")
            
            try:
                self.pipeline.process(kobj)
            finally:
                self.kobj_queue.task_done()
            logger.debug("Done")
    
    def kobj_processor_worker(self, timeout=0.1):
        while True:
            try:
                kobj = self.kobj_queue.get(timeout=timeout)
                logger.debug(f"Dequeued {kobj!r}")
                
                try:
                    self.pipeline.process(kobj)
                finally:
                    self.kobj_queue.task_done()
                logger.debug("Done")
            
            except queue.Empty:
                pass
            
            except Exception as e:
                logger.warning(f"Error processing kobj: {e}")
        
    def handle(
        self,
        rid: RID | None = None,
        manifest: Manifest | None = None,
        bundle: Bundle | None = None,
        event: Event | None = None,
        kobj: KnowledgeObject | None = None,
        event_type: EventType | None = None,
        source: KoiNetNode | None = None
    ):
        """Queues provided knowledge to be handled by processing pipeline.
        
        Knowledge may take the form of an RID, manifest, bundle, event, or knowledge object (with an optional event type for RID, manifest, or bundle objects). All objects will be normalized into knowledge objects and queued. If `flush` is `True`, the queue will be flushed immediately after adding the new knowledge.
        """
        if rid:
            _kobj = KnowledgeObject.from_rid(rid, event_type, source)
        elif manifest:
            _kobj = KnowledgeObject.from_manifest(manifest, event_type, source)
        elif bundle:
            _kobj = KnowledgeObject.from_bundle(bundle, event_type, source)
        elif event:
            _kobj = KnowledgeObject.from_event(event, source)
        elif kobj:
            _kobj = kobj
        else:
            raise ValueError("One of 'rid', 'manifest', 'bundle', 'event', or 'kobj' must be provided")
        
        self.kobj_queue.put(_kobj)
        logger.debug(f"Queued {_kobj!r}")



import logging
from functools import wraps

import cryptography.exceptions
from rid_lib.ext import Bundle
from rid_lib.ext.utils import sha256_hash
from .identity import NodeIdentity
from .protocol.envelope import UnsignedEnvelope, SignedEnvelope
from .protocol.secure import PublicKey
from .protocol.api_models import EventsPayload
from .protocol.event import EventType
from .protocol.node import NodeProfile
from .protocol.secure import PrivateKey
from .protocol.errors import (
    UnknownNodeError,
    InvalidKeyError,
    InvalidSignatureError,
    InvalidTargetError
)
from .effector import Effector
from .config import NodeConfig

logger = logging.getLogger(__name__)


class Secure:
    identity: NodeIdentity
    effector: Effector
    config: NodeConfig
    priv_key: PrivateKey
    
    def __init__(
        self, 
        identity: NodeIdentity, 
        effector: Effector,
        config: NodeConfig
    ):
        self.identity = identity
        self.effector = effector
        self.config = config

        self.priv_key = self._load_priv_key()
        
    def _load_priv_key(self) -> PrivateKey:
        with open(self.config.koi_net.private_key_pem_path, "r") as f:
            priv_key_pem = f.read()
        
        return PrivateKey.from_pem(
            priv_key_pem=priv_key_pem,
            password=self.config.env.priv_key_password
        )
        
    def _handle_unknown_node(self, envelope: SignedEnvelope) -> Bundle | None:
        if type(envelope.payload) != EventsPayload:
            return None
            
        for event in envelope.payload.events:
            # must be NEW event for bundle of source node's profile
            if event.rid != envelope.source_node:
                continue
            if event.event_type != EventType.NEW:
                continue            
            
            return event.bundle
        return None
        
    def create_envelope(self, payload, target) -> SignedEnvelope:
        return UnsignedEnvelope(
            payload=payload,
            source_node=self.identity.rid,
            target_node=target
        ).sign_with(self.priv_key)
        
    def validate_envelope(self, envelope: SignedEnvelope):
        node_bundle = (
            self.effector.deref(envelope.source_node) or
            self._handle_unknown_node(envelope)
        )
        
        if not node_bundle:
            raise UnknownNodeError(f"Couldn't resolve {envelope.source_node}")
        
        node_profile = node_bundle.validate_contents(NodeProfile)
        
        # check that public key matches source node RID
        if envelope.source_node.hash != sha256_hash(node_profile.public_key):
            raise InvalidKeyError("Invalid public key on new node!")
        
        # check envelope signed by validated public key
        pub_key = PublicKey.from_der(node_profile.public_key)
        try:
            envelope.verify_with(pub_key)
        except cryptography.exceptions.InvalidSignature as err:
            raise InvalidSignatureError(f"Signature {envelope.signature} is invalid.")
        
        # check that this node is the target of the envelope
        if envelope.target_node != self.identity.rid:
            raise InvalidTargetError(f"Envelope target {envelope.target_node!r} is not me")
        
    def envelope_handler(self, func):
        @wraps(func)
        async def wrapper(req: SignedEnvelope, *args, **kwargs) -> SignedEnvelope | None:
            logger.info("Validating envelope")
            
            self.validate_envelope(req)            
            logger.info("Calling endpoint handler")
            
            result = await func(req, *args, **kwargs)            
            
            if result is not None:
                logger.info("Creating response envelope")
                return self.create_envelope(
                    payload=result,
                    target=req.source_node
                )
        return wrapper




from logging import getLogger
from rid_lib.types import KoiNetNode
from ..protocol.event import Event, EventType
from ..identity import NodeIdentity
from ..effector import Effector
from ..network.event_queue import NetworkEventQueue

logger = getLogger(__name__)


class Actor:
    identity: NodeIdentity
    effector: Effector
    event_queue: NetworkEventQueue
    
    def __init__(
        self, 
        identity: NodeIdentity, 
        effector: Effector,
        event_queue: NetworkEventQueue
    ):
        self.identity = identity
        self.effector = effector
        self.event_queue = event_queue
    
    def handshake_with(self, target: KoiNetNode):
        logger.debug(f"Initiating handshake with {target}")
        self.event_queue.push_event_to(
            Event.from_rid(
                event_type=EventType.FORGET, 
                rid=self.identity.rid),
            node=target
        )
            
        self.event_queue.push_event_to(
            event=Event.from_bundle(
                event_type=EventType.NEW, 
                bundle=self.effector.deref(self.identity.rid)),
            node=target
        )
        
        self.event_queue.flush_webhook_queue(target)


import logging

from .network.behavior import Actor
from .effector import Effector
from .config import NodeConfig
from .processor.interface import ProcessorInterface
from .network.graph import NetworkGraph
from .identity import NodeIdentity

logger = logging.getLogger(__name__)


class NodeLifecycle:
    config: NodeConfig
    graph: NetworkGraph
    processor: ProcessorInterface
    effector: Effector
    actor: Actor
    
    def __init__(
        self,
        config: NodeConfig,
        identity: NodeIdentity,
        graph: NetworkGraph,
        processor: ProcessorInterface,
        effector: Effector,
        actor: Actor,
        use_kobj_processor_thread: bool
    ):
        self.config = config
        self.identity = identity
        self.graph = graph
        self.processor = processor
        self.effector = effector
        self.actor = actor
        self.use_kobj_processor_thread = use_kobj_processor_thread
    
    def start(self) -> None:
        """Starts a node, call this method first.
        
        Starts the processor thread (if enabled). Loads event queues into memory. Generates network graph from nodes and edges in cache. Processes any state changes of node bundle. Initiates handshake with first contact (if provided) if node doesn't have any neighbors.
        """
        if self.use_kobj_processor_thread:
            logger.info("Starting processor worker thread")
            self.processor.worker_thread.start()
        
        self.graph.generate()
        
        # refresh to reflect changes (if any) in config.yaml
        self.effector.deref(self.identity.rid, refresh_cache=True)
                
        logger.debug("Waiting for kobj queue to empty")
        if self.use_kobj_processor_thread:
            self.processor.kobj_queue.join()
        else:
            self.processor.flush_kobj_queue()
        logger.debug("Done")
    
        if not self.graph.get_neighbors() and self.config.koi_net.first_contact.rid:
            logger.debug(f"I don't have any neighbors, reaching out to first contact {self.config.koi_net.first_contact.rid!r}")
            
            self.actor.handshake_with(self.config.koi_net.first_contact.rid)
            
                        
    def stop(self):
        """Stops a node, call this method last.
        
        Finishes processing knowledge object queue. Saves event queues to storage.
        """
        logger.info("Stopping node...")
        
        if self.use_kobj_processor_thread:
            logger.info(f"Waiting for kobj queue to empty ({self.processor.kobj_queue.unfinished_tasks} tasks remaining)")
            self.processor.kobj_queue.join()
        else:
            self.processor.flush_kobj_queue()



from koi_net.effector import Effector
from rid_lib.ext import Cache
from .network.graph import NetworkGraph
from .network.event_queue import NetworkEventQueue
from .network.request_handler import RequestHandler
from .identity import NodeIdentity
from .processor.interface import ProcessorInterface


class ActionContext:
    identity: NodeIdentity
    effector: Effector

    def __init__(
        self,
        identity: NodeIdentity,
        effector: Effector
    ):
        self.identity = identity
        self.effector = effector
    

class HandlerContext:
    identity: NodeIdentity
    cache: Cache
    event_queue: NetworkEventQueue
    graph: NetworkGraph
    request_handler: RequestHandler
    effector: Effector
    _processor: ProcessorInterface | None
    
    def __init__(
        self,
        identity: NodeIdentity,
        cache: Cache,
        event_queue: NetworkEventQueue,
        graph: NetworkGraph,
        request_handler: RequestHandler,
        effector: Effector
    ):
        self.identity = identity
        self.cache = cache
        self.event_queue = event_queue
        self.graph = graph
        self.request_handler = request_handler
        self.effector = effector
        self._processor = None
        
    def set_processor(self, processor: ProcessorInterface):
        self._processor = processor
        
    @property
    def handle(self):
        return self._processor.handle


import logging
from base64 import urlsafe_b64decode, urlsafe_b64encode
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives import serialization
from rid_lib.ext.utils import sha256_hash

logger = logging.getLogger(__name__)


class PrivateKey:
    priv_key: ec.EllipticCurvePrivateKey
    
    def __init__(self, priv_key):
        self.priv_key = priv_key
        
    @classmethod
    def generate(cls):
        return cls(priv_key=ec.generate_private_key(ec.SECP192R1()))

    def public_key(self) -> "PublicKey":
        return PublicKey(self.priv_key.public_key())
    
    @classmethod
    def from_pem(cls, priv_key_pem: str, password: str):
        return cls(
            priv_key=serialization.load_pem_private_key(
                data=priv_key_pem.encode(),
                password=password.encode()
            )
        )

    def to_pem(self, password: str) -> str:
        return self.priv_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.BestAvailableEncryption(password.encode())
        ).decode()
        
    def sign(self, message: bytes) -> str:
        hashed_message = sha256_hash(message.decode())
        
        signature = urlsafe_b64encode(
            self.priv_key.sign(
                data=message,
                signature_algorithm=ec.ECDSA(hashes.SHA256())
            )
        ).decode()
        
        logger.debug(f"Signing message with [{self.public_key().to_der()}]")
        logger.debug(f"hash: {hashed_message}")
        logger.debug(f"signature: {signature}")
        
        return signature
                

class PublicKey:
    pub_key: ec.EllipticCurvePublicKey
    
    def __init__(self, pub_key):
        self.pub_key = pub_key
    
    @classmethod
    def from_pem(cls, pub_key_pem: str):
        return cls(
            pub_key=serialization.load_pem_public_key(
                data=pub_key_pem.encode()
            )
        )
        
    def to_pem(self) -> str:
        return self.pub_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        ).decode()
        
    @classmethod
    def from_der(cls, pub_key_der: str):        
        return cls(
            pub_key=serialization.load_der_public_key(
                data=urlsafe_b64decode(pub_key_der)
            )
        )
    
    def to_der(self) -> str:
        return urlsafe_b64encode(
            self.pub_key.public_bytes(
                encoding=serialization.Encoding.DER,
                format=serialization.PublicFormat.SubjectPublicKeyInfo
            )
        ).decode()
        
    def verify(self, signature: str, message: bytes) -> bool:
        hashed_message = sha256_hash(message.decode())

        logger.debug(f"Verifying message with [{self.to_der()}]")
        logger.debug(f"hash: {hashed_message}")
        logger.debug(f"signature: {signature}")
        
        # NOTE: throws cryptography.exceptions.InvalidSignature on failure
        
        self.pub_key.verify(
            signature=urlsafe_b64decode(signature),
            data=message,
            signature_algorithm=ec.ECDSA(hashes.SHA256())
        )



import logging
import httpx
from rid_lib import RID
from rid_lib.core import RIDType
from rid_lib.ext import Cache, Bundle
from rid_lib.types import KoiNetNode

from .graph import NetworkGraph
from .request_handler import RequestHandler
from ..protocol.node import NodeProfile, NodeType
from ..protocol.event import Event
from ..protocol.api_models import ErrorResponse
from ..identity import NodeIdentity
from ..config import NodeConfig
from ..effector import Effector

logger = logging.getLogger(__name__)


class NetworkResolver:
    """A collection of functions and classes to interact with the KOI network."""
    
    config: NodeConfig    
    identity: NodeIdentity
    effector: Effector
    cache: Cache
    graph: NetworkGraph
    request_handler: RequestHandler
    
    def __init__(
        self, 
        config: NodeConfig,
        cache: Cache, 
        identity: NodeIdentity,
        effector: Effector,
        graph: NetworkGraph,
        request_handler: RequestHandler,
    ):
        self.config = config
        self.identity = identity
        self.cache = cache
        self.graph = graph
        self.request_handler = request_handler
        self.effector = effector
        
        self.poll_event_queue = dict()
        self.webhook_event_queue = dict()
    
    def get_state_providers(self, rid_type: RIDType) -> list[KoiNetNode]:
        """Returns list of node RIDs which provide state for the specified RID type."""
        
        logger.debug(f"Looking for state providers of '{rid_type}'")
        provider_nodes = []
        for node_rid in self.cache.list_rids(rid_types=[KoiNetNode]):
            if node_rid == self.identity.rid:
                continue
            
            node_bundle = self.cache.read(node_rid)
            
            node_profile = node_bundle.validate_contents(NodeProfile)
            
            if (node_profile.node_type == NodeType.FULL) and (rid_type in node_profile.provides.state):
                logger.debug(f"Found provider {node_rid!r}")
                provider_nodes.append(node_rid)
        
        if not provider_nodes:
            logger.debug("Failed to find providers")
        return provider_nodes
            
    def fetch_remote_bundle(self, rid: RID) -> tuple[Bundle | None, KoiNetNode | None]:
        """Attempts to fetch a bundle by RID from known peer nodes."""
        
        logger.debug(f"Fetching remote bundle {rid!r}")
        remote_bundle, node_rid = None, None
        for node_rid in self.get_state_providers(type(rid)):
            payload = self.request_handler.fetch_bundles(
                node=node_rid, rids=[rid])
            
            if payload.bundles:
                remote_bundle = payload.bundles[0]
                logger.debug(f"Got bundle from {node_rid!r}")
                break
        
        if not remote_bundle:
            logger.warning("Failed to fetch remote bundle")
            
        return remote_bundle, node_rid
    
    def fetch_remote_manifest(self, rid: RID) -> tuple[Bundle | None, KoiNetNode | None]:
        """Attempts to fetch a manifest by RID from known peer nodes."""
        
        logger.debug(f"Fetching remote manifest {rid!r}")
        remote_manifest, node_rid = None, None
        for node_rid in self.get_state_providers(type(rid)):
            payload = self.request_handler.fetch_manifests(
                node=node_rid, rids=[rid])
            
            if payload.manifests:
                remote_manifest = payload.manifests[0]
                logger.debug(f"Got bundle from {node_rid!r}")
                break
        
        if not remote_manifest:
            logger.warning("Failed to fetch remote bundle")
            
        return remote_manifest, node_rid
    
    def poll_neighbors(self) -> dict[KoiNetNode, list[Event]]:
        """Polls all neighboring nodes and returns compiled list of events.
        
        If this node has no neighbors, it will instead attempt to poll the provided first contact URL.
        """
        
        graph_neighbors = self.graph.get_neighbors()
        neighbors = []
        
        if graph_neighbors:
            for node_rid in graph_neighbors:
                node_bundle = self.cache.read(node_rid)
                if not node_bundle: 
                    continue
                node_profile = node_bundle.validate_contents(NodeProfile)
                if node_profile.node_type != NodeType.FULL: 
                    continue
                neighbors.append(node_rid)
            
        elif self.config.koi_net.first_contact.rid:
            neighbors.append(self.config.koi_net.first_contact.rid)
        
        event_dict = dict()
        for node_rid in neighbors:
            try:
                payload = self.request_handler.poll_events(
                    node=node_rid, 
                    rid=self.identity.rid
                )
                
                if type(payload) == ErrorResponse:
                    continue
                    
                if payload.events:
                    logger.debug(f"Received {len(payload.events)} events from {node_rid!r}")
                    
                    event_dict[node_rid] = payload.events
                    
            except httpx.ConnectError:
                logger.debug(f"Failed to reach node {node_rid!r}")
                continue
        
        return event_dict


import logging
from queue import Queue
import httpx
from pydantic import BaseModel
from rid_lib import RID
from rid_lib.ext import Cache
from rid_lib.types import KoiNetNode

from .graph import NetworkGraph
from .request_handler import RequestHandler
from ..protocol.node import NodeProfile, NodeType
from ..protocol.edge import EdgeProfile, EdgeType
from ..protocol.event import Event
from ..identity import NodeIdentity
from ..config import NodeConfig
from ..effector import Effector

logger = logging.getLogger(__name__)


class EventQueueModel(BaseModel):
    webhook: dict[KoiNetNode, list[Event]]
    poll: dict[KoiNetNode, list[Event]]

type EventQueue = dict[RID, Queue[Event]]

class NetworkEventQueue:
    """A collection of functions and classes to interact with the KOI network."""
    
    config: NodeConfig    
    identity: NodeIdentity
    effector: Effector
    cache: Cache
    graph: NetworkGraph
    request_handler: RequestHandler
    poll_event_queue: EventQueue
    webhook_event_queue: EventQueue
    
    def __init__(
        self, 
        config: NodeConfig,
        cache: Cache, 
        identity: NodeIdentity,
        effector: Effector,
        graph: NetworkGraph,
        request_handler: RequestHandler,
    ):
        self.config = config
        self.identity = identity
        self.cache = cache
        self.graph = graph
        self.request_handler = request_handler
        self.effector = effector
        
        self.poll_event_queue = dict()
        self.webhook_event_queue = dict()
    
    def _load_event_queues(self):
        """Loads event queues from storage."""
        try:
            with open(self.config.koi_net.event_queues_path, "r") as f:
                queues = EventQueueModel.model_validate_json(f.read())
            
            for node in queues.poll.keys():
                for event in queues.poll[node]:
                    queue = self.poll_event_queue.setdefault(node, Queue())
                    queue.put(event)
            
            for node in queues.webhook.keys():
                for event in queues.webhook[node]:
                    queue = self.webhook_event_queue.setdefault(node, Queue())
                    queue.put(event)
                                
        except FileNotFoundError:
            return
        
    def _save_event_queues(self):
        """Writes event queues to storage."""
        events_model = EventQueueModel(
            poll={
                node: list(queue.queue) 
                for node, queue in self.poll_event_queue.items()
                if not queue.empty()
            },
            webhook={
                node: list(queue.queue) 
                for node, queue in self.webhook_event_queue.items()
                if not queue.empty()
            }
        )
        
        if len(events_model.poll) == 0 and len(events_model.webhook) == 0:
            return
        
        with open(self.config.koi_net.event_queues_path, "w") as f:
            f.write(events_model.model_dump_json(indent=2))
    
    def push_event_to(self, event: Event, node: KoiNetNode, flush=False):
        """Pushes event to queue of specified node.
        
        Event will be sent to webhook or poll queue depending on the node type and edge type of the specified node. If `flush` is set to `True`, the webhook queued will be flushed after pushing the event.
        """
        logger.debug(f"Pushing event {event.event_type} {event.rid!r} to {node}")
                        
        node_bundle = self.effector.deref(node)
        
        # if there's an edge from me to the target node, override broadcast type
        edge_rid = self.graph.get_edge(
            source=self.identity.rid,
            target=node
        )
        
        edge_bundle = self.effector.deref(edge_rid) if edge_rid else None
        
        if edge_bundle:
            logger.debug(f"Found edge from me to {node!r}")
            edge_profile = edge_bundle.validate_contents(EdgeProfile)
            if edge_profile.edge_type == EdgeType.WEBHOOK:
                event_queue = self.webhook_event_queue
            elif edge_profile.edge_type == EdgeType.POLL:
                event_queue = self.poll_event_queue
                
        elif node_bundle:
            logger.debug(f"Found bundle for {node!r}")
            node_profile = node_bundle.validate_contents(NodeProfile)
            if node_profile.node_type == NodeType.FULL:
                event_queue = self.webhook_event_queue
            elif node_profile.node_type == NodeType.PARTIAL:
                event_queue = self.poll_event_queue
        
        elif node == self.config.koi_net.first_contact.rid:
            logger.debug(f"Node {node!r} is my first contact")
            # first contact node is always a webhook node
            event_queue = self.webhook_event_queue
        
        else:
            logger.warning(f"Node {node!r} unknown to me")
            return
        
        queue = event_queue.setdefault(node, Queue())
        queue.put(event)
                
        if flush and event_queue is self.webhook_event_queue:
            self.flush_webhook_queue(node)
            
    def _flush_queue(self, event_queue: EventQueue, node: KoiNetNode) -> list[Event]:
        """Flushes a node's queue, returning list of events."""
        queue = event_queue.get(node)
        events = list()
        if queue:
            while not queue.empty():
                event = queue.get()
                logger.debug(f"Dequeued {event.event_type} {event.rid!r}")
                events.append(event)
        
        return events
    
    def flush_poll_queue(self, node: KoiNetNode) -> list[Event]:
        """Flushes a node's poll queue, returning list of events."""
        logger.debug(f"Flushing poll queue for {node}")
        return self._flush_queue(self.poll_event_queue, node)
    
    def flush_webhook_queue(self, node: KoiNetNode, requeue_on_fail: bool = True):
        """Flushes a node's webhook queue, and broadcasts events.
        
        If node profile is unknown, or node type is not `FULL`, this operation will fail silently. If the remote node cannot be reached, all events will be requeued.
        """
        
        logger.debug(f"Flushing webhook queue for {node}")
        
        # node_bundle = self.effector.deref(node)
        
        # if not node_bundle:
        #     logger.warning(f"{node!r} not found")
        #     return
        
        # node_profile = node_bundle.validate_contents(NodeProfile)
        
        # if node_profile.node_type != NodeType.FULL:
        #     logger.warning(f"{node!r} is a partial node!")
        #     return
        
        events = self._flush_queue(self.webhook_event_queue, node)
        if not events: return
        
        logger.debug(f"Broadcasting {len(events)} events")
        
        try:  
            self.request_handler.broadcast_events(node, events=events)
            return True
        except httpx.ConnectError:
            logger.warning("Broadcast failed")
            
            if requeue_on_fail:
                for event in events:
                    self.push_event_to(event, node)
            return False



import logging
from pydantic import Field
from rich.logging import RichHandler
from koi_net import NodeInterface
from koi_net.protocol.node import NodeProfile, NodeType
from koi_net.config import NodeConfig, KoiNetConfig

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    handlers=[RichHandler()]
)

logging.getLogger("koi_net").setLevel(logging.DEBUG)
logger = logging.getLogger(__name__)


class PartialNodeConfig(NodeConfig):
    koi_net: KoiNetConfig | None = Field(default_factory = lambda:
        KoiNetConfig(
            node_name="partial",
            node_profile=NodeProfile(
                node_type=NodeType.PARTIAL
            ),
            cache_directory_path=".partial_rid_cache",
            event_queues_path="partial_event_queues.json",
            private_key_pem_path="partial_priv_key.pem"
        )
    )

node = NodeInterface(
    config=PartialNodeConfig.load_from_yaml("partial_config.yaml")
)

if __name__ == "__main__":
    node.poller.run()